{% extends main.html %}
{% block content %}

  <h1 class="title"> Scheduler {{scheduler.address}} </h1>

  <a class="button is-primary" href="logs.html">Logs</a>
  <a class="button is-primary" href="../../status">Bokeh</a>

  {% set worker_list = list(workers.values()) %}
  {% include "worker-table.html" %}

{% end %}
